# -*- coding: utf-8 -*-
"""pytest to see if extras_require worked.

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function

def test_import():
    import adhan
    import civicjson
