# -*- coding: utf-8 -*-
"""Test that module imports.

You should delete the test once you have real tests.
Only necessary if you have no other tests so that
tox will work.
"""
import pytest


def test_1():
    import private
