"""Test that module imports.

You should delete the test once you have real tests.
Only necessary if you have no other tests so that
pykern.pkcli.test passes.
"""

def test_import():
    import {{ name }}
