# Read The Docs configuration file
# See https://docs.readthedocs.io/en/stable/config-file/v2.html for details
version: 2
build:
  os: ubuntu-22.04
  tools:
    python: "3.9"
  jobs:
    pre_build:
     - curl https://radia.run | bash -s readthedocs
sphinx:
  configuration: docs/conf.py
