Welcome to {{ name }}
=================

{{ description }}

.. autosummary::
   :toctree: _autosummary
   :recursive:

.. toctree::
   :maxdepth: 2

   {{ name }}
   modules

Indices and tables
==================

* :ref:`genindex`
* :ref:`modindex`
* :ref:`search`
