def force_arg_per_line(
    this,
    be,
    a,
    very,
    poorly,
    formatted,
    function,
    that,
    noone,
    should,
    ever,
    use,
    it,
    has,
    way,
    too,
    many,
    arguments,
):
    foo("force_arg_per_line")
