"""sleeps so other tests can run

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_sleep():
    import time

    time.sleep(1)
